#!/usr/bin/env python

## application indicator (ubuntu >= 11.04 ?) as UI for samsung scanner server

## version 0.4.5 by totally-king

import sys
import os
import time
import re
import subprocess
import threading
import signal
import pwd
from string import Template

from imp import load_source

from gi.repository import Gtk, Gdk, GObject
from gi.repository import AppIndicator3 as appindicator
from gi.repository import Notify
from gi.repository import Wnck


configuration = load_source('configuration', '/etc/samsungScannerServer.conf')
LOG_FILENAME = configuration.LOG_NAME
if hasattr(configuration, 'OWNER_UID'):
    OWNER_UID = configuration.OWNER_UID
else:
    OWNER_UID = 1000
HOME_DIR = pwd.getpwuid(OWNER_UID).pw_dir
OUTPUT_PREFIX = configuration.OUTPUT_PREFIX
SCAN_FOLDER = '/'.join(Template(OUTPUT_PREFIX).safe_substitute(homedir=HOME_DIR).split('/')[:-1])
SCAN_FOLDER_LAST = SCAN_FOLDER.split('/')[-1]



############################ init notify-osd ###########################

# even in Python this is globally nasty :), do something nicer in your own code
capabilities = {'actions':                 False,
        'body':                            False,
        'body-hyperlinks':                 False,
        'body-images':                     False,
        'body-markup':                     False,
        'icon-multi':                      False,
        'icon-static':                     False,
        'sound':                           False,
        'image/svg+xml':                   False,
        'x-canonical-private-synchronous': False,
        'x-canonical-append':              False,
        'x-canonical-private-icon-only':   False,
        'x-canonical-truncation':          False}

def initCaps ():
    caps = Notify.get_server_caps ()
    if caps is None:
        sys.stderr.write("Failed to receive server caps.\n")
        sys.exit (1)

    for cap in caps:
        capabilities[cap] = True

def printCaps ():
    info = Notify.get_server_info ()
    print "Name:          " + info[1]
    print "Vendor:        " + info[2]
    print "Version:       " + info[3]
    print "Spec. Version: " + info[4]

    caps = Notify.get_server_caps ()
    if caps is None:
        sys.stderr.write("Failed to receive server caps.\n")
        sys.exit (1)

    print "Supported capabilities/hints:"
    if capabilities['actions']:
        print "\tactions"
    if capabilities['body']:
        print "\tbody"
    if capabilities['body-hyperlinks']:
        print "\tbody-hyperlinks"
    if capabilities['body-images']:
        print "\tbody-images"
    if capabilities['body-markup']:
        print "\tbody-markup"
    if capabilities['icon-multi']:
        print "\ticon-multi"
    if capabilities['icon-static']:
        print "\ticon-static"
    if capabilities['sound']:
        print "\tsound"
    if capabilities['image/svg+xml']:
        print "\timage/svg+xml"
    if capabilities['x-canonical-private-synchronous']:
        print "\tx-canonical-private-synchronous"
    if capabilities['x-canonical-append']:
        print "\tx-canonical-append"
    if capabilities['x-canonical-private-icon-only']:
        print "\tx-canonical-private-icon-only"
    if capabilities['x-canonical-truncation']:
        print "\tx-canonical-truncation"

    print "Notes:"
    if info[1] == "notify-osd":
        print "\tx- and y-coordinates hints are ignored"
        print "\texpire-timeout is ignored"
        print "\tbody-markup is accepted but filtered"
    else:
        print "\tnone"

def pushNotification (body, title='Samsung Scanner Server', icon='scannerServer', new=False):
    n = Notify.Notification.new (title, body, icon)
    ## if the system can and no new is wanted, allow merging of notifications
    if capabilities['x-canonical-append'] and not new:
        n.set_hint_string ('x-canonical-append', 'allowed')
    n.show ()



############################### FUNCTIONS ###############################

def followLog(filename, printing=False, nrLinesAtStart=0, sleepingTime=0.1, yieldNone=False):
    '''thefile: file object with log file
       generator -> yields one new line at the end of the log file at a time'''
        
    def getLinesAtStart(nrLinesAtStart):
        ## buffer size of blocks the file is read in
        BUFSIZ = 1024
        thefile.seek(0, os.SEEK_END)
        fsize = thefile.tell()
        block = -1
        data = ''
        exitLoop = False
        while not exitLoop:
            step = (block * BUFSIZ)
            if abs(step) >= fsize:
                ## cannot read more than file, so start at 0
                thefile.seek(0)
                ## and do not read complete BUFSIZ but what is remaining
                newdata = thefile.read(BUFSIZ-(abs(step)-fsize))
                ## finish (no more left to read)
                exitLoop = True
            else:
                ## go step size ahead of end
                thefile.seek(step, os.SEEK_END)
                ## read BUFSIZ sized block
                newdata = thefile.read(BUFSIZ)
            ## remove final \n of first block read
            if data == '':
                newdata = newdata.rstrip('\n')
            ## sew together with what was read in before
            data = newdata + data
            ## finish if we get enough lines
            if data.count('\n') >= nrLinesAtStart:
                break
            else:
                block -= 1
        return data.splitlines()[-nrLinesAtStart:]
    
    i = 0
    while not os.path.exists(filename):
        if i % 10 == 0:
            print "No log file '%s' found, retrying ..." % filename
        i += 1
        time.sleep(.05)
    
    thefile = open(filename, 'r')
    
    if bool(nrLinesAtStart):
        for line in getLinesAtStart(nrLinesAtStart):
            yield line + '\n'
   
    thefile.seek(0, os.SEEK_END) ## move to end of the file

    filename = thefile.name
    fileSizeOld = 0
    while True:
        ## get current log file size
        try:
            fileSize = os.stat(filename).st_size
        except OSError:
            fileSize = 0
        ## if log rotated
        if fileSize < fileSizeOld:
            ## empty out end of old log file
            for line in thefile:
                yield line
            if printing:
                print '='*23 + ' log rotated ' + '='*23
            ## then close and reopen to get new log file
            thefile.close()
            while True:
                try:
                    thefile = open(filename, 'r')
                except OSError:
                    time.sleep(sleepingTime)
                    continue
                else:
                    break
        ## then continue as before
        line = thefile.readline()
        if not line:
            time.sleep(sleepingTime) ## sleep briefly
            ## yields None to make a non-blocking call to this generator possible
            if yieldNone:
                yield None
            continue
        yield line
        ## make 'current' log file size the old one
        fileSizeOld = fileSize

def parseLogLine(rawline):
    '''some output formatting for parsing log lines; return only
       messages, not time etc.'''
    line = rawline.rstrip('\n')
    linenotabs = line
    for i in range(5):
        linenotabs = linenotabs.replace('  ', ' ')
    try:
        dateRow = linenotabs.split(' ')[0]
        timeRow = linenotabs.split(' ')[1]
        if re.match('^\d\d-\d\d-\d\d$', dateRow) and \
           re.match('^\d\d:\d\d:\d\d$', timeRow):
            msg = ' '.join(linenotabs.split(' ')[4:])
        else:
            msg = line
    except IndexError:
        msg = line
    return msg

def filterLogMsg(msg):
    '''let thru only log messages that are error-related, remove traceback'''
    if msg[0] == '#' or re.match('Used .* as configuration file', msg) or \
      msg == 'Below is what was configured with it.' or \
      msg.startswith('CONFIG:') or msg.startswith('AUTOCONFIG:') or \
      msg == 'The following was automatically configured.' or \
      msg == 'Init SANE ...' or \
      msg == 'Getting server IP and setting scanner name so that SANE uses proxy.' or \
      msg == 'Connecting to scanner ...' or msg == 'Connected to scanner.' or \
      msg.endswith('cache reset.') or \
      msg.startswith('Write PID to file:') or msg.startswith('At program termination') or \
      msg.startswith('<function') or msg.startswith('InstanceID') or \
      msg.startswith('TCProxy:') or msg.startswith('UDProxy:') or \
      msg.startswith('Newly registered server') or \
      msg.startswith('Caught signal') or msg.startswith('Unregistered server') or \
      msg.startswith('Could not remove PID file:') or msg.startswith('because it was already deleted') or \
      msg.startswith('Removed PID file:') or \
      msg.startswith('Traceback') or msg.startswith('  '):
        return None
    else:
        return msg

def serverRunning():
    ps = subprocess.Popen(['ps', 'aux'], stdout=subprocess.PIPE)
    out, err = ps.communicate()
    pids = []
    for process in out.split('\n'):
        if re.match('.*python.*ScannerServer\.py', process):
            pid = process.split()[1]
            pids.append(pid)
    return pids

def spawnDaemon(func, *args, **kwargs):
    # do the UNIX double-fork magic, see Stevens' "Advanced 
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    try: 
        pid = os.fork() 
        if pid > 0:
            # parent process, return and keep running
            return
    except OSError, e:
        print >>sys.stderr, "fork #1 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1)
    os.setsid()
    # do second fork
    try: 
        pid = os.fork() 
        if pid > 0:
            # exit from second parent
            os._exit(os.EX_OK)
            #sys.exit(0) 
    except OSError, e: 
        print >>sys.stderr, "fork #2 failed: %d (%s)" % (e.errno, e.strerror) 
        sys.exit(1)
    # do stuff
    func(*args, **kwargs)
    # all done
    os._exit(os.EX_OK)

def server(action):
    spawnDaemon(os.system, 'sudo service samsungScannerServer %s' % action)

def wfcCleanupAndStart(what='started', notify=True):
    '''cleanup existing WFC threads and start a new one'''
    global WFC_INSTANCES
    for old_wfc in WFC_INSTANCES:
        if old_wfc.is_alive():
            old_wfc.join()
    WFC_INSTANCES = []
    new_wfc = WaitForConnection(what=what, notify=notify)
    new_wfc.start()
    WFC_INSTANCES.append(new_wfc)

def startServer(notify=True):
    if serverRunning():
        restartServer(notify=notify)
    else:
        server('start')
        wfcCleanupAndStart()

def stopServer(waitTime=60, notify=True):
    server('stop')
    ## wait for it to stop, max. waitTime in seconds
    i = 0
    while True:
        pids = serverRunning()
        if not pids:
            if notify:
                pushNotification('stopped ...')
            return 0
        if i < waitTime:
            time.sleep(1)
            i += 1
            continue
        else:
            pushNotification("could not be stopped.\nTry 'sudo kill %s'!" % ' '.join(pids) )
            return 1

def restartServer(notify=True):
    if serverRunning():
        stopServer(notify=False)
        server('start')
        wfcCleanupAndStart(what='restarted')
    else:
        startServer(notify=notify)

def checkStatus():
    if serverRunning():
        wfcCleanupAndStart(what=None)
    else:
        pushNotification('is not running ...')

def stopAndExit(notify=True):
    Gtk.main_quit()
    stopServer(notify=notify)
    if 'TERMINAL' in globals():
        TERMINAL.terminate()
        TERMINAL.wait()
    sys.exit(0)

def logWatcher():
    ## get the name of this file without extension
    scriptname = '.'.join( os.path.basename(__file__).split('.')[:-1] )
    call = ['gnome-terminal', '--title=Samsung Scanner Server Log File',
            '--window', '--maximize',
            '''--command=python -c "
from %(scriptname)s import followLog, LOG_FILENAME
logLines = followLog(LOG_FILENAME, printing=True, nrLinesAtStart=100)
for rawline in logLines:
    print rawline,
            "''' % locals()]
    global TERMINAL
    TERMINAL = subprocess.Popen(call)

def openScansFolder():
    ## check if nautilus scan folder is already open
    openFlag = False
    screen = Wnck.Screen.get_default()
    screen.force_update()
    for w in screen.get_windows_stacked():
        if w.get_name() == SCAN_FOLDER_LAST and \
           w.get_application().get_name() == 'nautilus':
            openFlag = True
            w.activate(int(time.time()))
            break
    if not openFlag:
        call = ['nautilus', SCAN_FOLDER]
        subprocess.call(call)
    ind.set_status(appindicator.IndicatorStatus.ACTIVE)
    WL.reset_newFiles_cnt()
    menu.get_children()[0].set_label(titles[0])


def buttonWrapper(menuItem, func, *args):
    func(*args)


################################ CLASSES ################################

class MyThread(threading.Thread):
    '''parent class for my threads'''
    LOG_FILENAME = LOG_FILENAME
    
    def __init__(self):
        super(MyThread, self).__init__()
        self._stoprequest = threading.Event()
    
    def join(self, timeout=None):
        self._stoprequest.set()
        super(MyThread, self).join(timeout)

    def pushNotification(self, *args, **kwargs):
        GObject.idle_add(pushNotification, *args, **kwargs)


class WaitForConnection(MyThread):
    '''a thread that waits for a connected server and issues notifications'''
    
    def __init__(self, what='started', notify=True):
        '''what = started / restarted
           what = None used for checking server status'''
        super(WaitForConnection, self).__init__()
        self.what = what
        self.notify = notify
    
    def run(self):
        if self.what: ## starting or restarting server
            logLines = followLog(self.LOG_FILENAME, yieldNone=True)
        else: ## checking server status
            logLines = followLog(self.LOG_FILENAME, nrLinesAtStart=1, yieldNone=True)
        while not self._stoprequest.is_set():
            rawline = logLines.next() ## yields None if there are no new lines
            if rawline:
                msg = filterLogMsg(parseLogLine(rawline))
                if msg:
                    if not re.match('Waiting for scan job', msg):
                        if self.what: ## starting or restarting server
                            self.pushNotification('cannot be %s: %s' % (self.what, msg) )
                        else: ## checking server status
                            self.pushNotification('%s' % (msg) )
                    else:
                        if self.notify:
                            if self.what: ## starting or restarting server
                                self.pushNotification("%s ... You can now use 'Scan to PC'." % self.what)
                            else: ## checking server status
                                self.pushNotification("is up and running ... You can use 'Scan to PC'.")
                        break


class WatchLog(MyThread):
    '''a thread that watches the log file for incoming scan files'''
    
    def __init__(self):
        super(WatchLog, self).__init__()
        self.newFiles = 0
        
    def ordinal(self, n):
        if 10 < n < 14: return '%sth' % n
        if n % 10 == 1: return '%sst' % n
        if n % 10 == 2: return '%snd' % n
        if n % 10 == 3: return '%srd' % n
        return '%sth' % n
    
    def _attention(self):
        ind.set_status(appindicator.IndicatorStatus.ATTENTION)
        ## modify first menu item (which is "Open Scans Directory")
        menu.get_children()[0].set_label(titles[0] + ' (%d new)' % self.newFiles)
        
    def attention(self):
        self.newFiles += 1
        GObject.idle_add(self._attention)
    
    def reset_newFiles_cnt(self):
        self.newFiles = 0
    
    def run(self):
        logLines = followLog(self.LOG_FILENAME, yieldNone=True)
        scanning, saving, savingFinal, multiPDF = False, False, False, False
        while not self._stoprequest.is_set():
            rawline = logLines.next() ## yields None if there are no new lines
            if rawline:
                msg = parseLogLine(rawline)
                if scanning:
                    if saving or savingFinal:
                        if msg == 'Done.':
                            if multiPDF:
                                if savingFinal:
                                    self.pushNotification('Multipage PDF file is done: %s' % filename)
                                    self.attention()
                                    scanning, multiPDF, savingFinal = False, False, False
                                else:
                                    pageCnt += 1
                                    pageCntOrd = self.ordinal(pageCnt)
                                    self.pushNotification('%s PDF page arrived.' % pageCntOrd)
                            else:
                                self.pushNotification('New %s file arrived: %s' % \
                                                        (filename.split('.')[-1].upper(), filename) )
                                self.attention()
                            saving, savingFinal = False, False
                            continue
                    matchSave = re.match('Saving (.*) \\.\\.\\.', msg)
                    if matchSave:
                        saving = True
                        filename = matchSave.group(1).split('/')[-1]
                        continue
                    if multiPDF:
                        matchFinal = re.match('Writing final PDF (.*) \\.\\.\\.', msg)
                        if matchFinal:
                            savingFinal = True
                            filename = matchFinal.group(1).split('/')[-1]
                            continue
                    if msg == 'TCProxy: "no more pages"':
                        if multiPDF and pageCnt == 1:
                            self.pushNotification('Final PDF file is done: %s' % filename)
                            self.attention()
                            scanning, multiPDF = False, False
                        elif not multiPDF:
                            scanning, multiPDF = False, False
                        continue
                if msg == 'Scanning ...':
                    self.pushNotification('Receiving scan ...')
                    scanning = True
                    continue
                matchOpt = re.match('Options selected by user: (.*)', msg)
                if matchOpt:
                    user_options = eval(matchOpt.group(1))
                    if user_options['format'] == 'FORMAT_M_PDF' or user_options['format'] == 'FORMAT_PDF':
                        multiPDF = True
                        pageCnt = 0
                    continue


################################# MAIN #################################

if __name__ == "__main__":
    ## initialize threading for GUI
    GObject.threads_init()
    ## initialize notifiy-osd stuff
    if not Notify.init ("samsungScannerServer"):
        sys.stderr.write("Could not initialize notifications.\n")
        sys.exit (1)
    ## call this so we can savely use capabilities dictionary later
    initCaps ()
    
    ## set up app indicator
    ind = appindicator.Indicator.new(
                        "samsungScannerServer",
                        "indicator-scannerServer",
                        appindicator.IndicatorCategory.APPLICATION_STATUS)
    ind.set_status(appindicator.IndicatorStatus.ACTIVE)
    ind.set_attention_icon("indicator-scannerServer-new")
    
    ## global list of WaitForConnection instances (there should only be one at a time)
    WFC_INSTANCES = []
    
    ## start samsung scanner server
    startServer()
    
    ## start thread that watches log file for incoming scan files
    WL = WatchLog()
    WL.start()
    
    ## create a menu
    menu = Gtk.Menu()
    
    ## create some menu items
    items = {}
    titles = []
    ## title = button name | dict value = list of [ Gtk.STOCK_..., tuple of (func, *args) ]
    title = '_Open Scans Directory';       titles.append(title)
    items[title] = [Gtk.STOCK_OPEN, (openScansFolder,)]
    title = 'Stop Server and _Exit';    titles.append(title)
    items[title] = [Gtk.STOCK_QUIT, (stopAndExit,)]
    title = 'Separator';    titles.append(title)
    title = '_Check Server Status';     titles.append(title)
    items[title] = [Gtk.STOCK_DIALOG_INFO, (checkStatus,)]
    title = '_Restart Server';          titles.append(title)
    items[title] = [Gtk.STOCK_REFRESH, (restartServer,)]
    title = '_Watch Log File';          titles.append(title)
    items[title] = [Gtk.STOCK_FIND, (logWatcher,)]
    del title
    
    for title in titles:
        if title == 'Separator':
            menuItem = Gtk.SeparatorMenuItem.new()
        else:
            menuItem = Gtk.ImageMenuItem.new_with_mnemonic(title)
            menuItem.set_image(Gtk.Image.new_from_stock(items[title][0], Gtk.IconSize.MENU))
            menuItem.set_always_show_image(True)
            ## menuItem.connect('activate', buttonWrapper, function[, arguments])
            menuItem.connect('activate', buttonWrapper, *items[title][1])
        
        menuItem.show()
        menu.append(menuItem)
    
    ind.set_menu(menu)
    
    ## update ImageMenuItem labels (which my have changed back to stock)
    i = 0
    for menuItem in menu.get_children():
        title = titles[i]
        if title != 'Separator':
            menuItem.set_label(title)
        i += 1
    del i, title
    
    Gtk.main()
